#pragma once

// micro platform-dependent sys core

class Mutex
{
protected:
    void *handle;		// system-specific mutex handle
public:
    Mutex();
    ~Mutex();

    void lock();
    void unlock();
};

class MutexLock
{
protected:
    Mutex *mref;		// mutex refptr
public:
    MutexLock( Mutex &m );
    ~MutexLock();
};

class Event
{
protected:
    void *sys;		// internal system handle
public:
    // if autoreset is true, event will be reset automatically
    Event( bool autoReset = false );
    ~Event();

    // signal(set) event
    void signal();
    // reset event
    void reset();

    // wait for this event (wait time given in milliseconds or waitForever constant)
    // returns true on success (=>event set)
    bool wait( int ms = -1 );
};

class Thread
{
protected:
    void *thrParam;
    void *handle;					// sys-specific handle
    volatile bool shouldTerminate;	// thread should terminate flag
    volatile bool shouldStart;      // should start on resume?
    Mutex mutex;                    // local thread mutex
public:
    Thread( void *param = 0 );
    virtual ~Thread();

    // resume thread
    virtual bool resume();
    // suspend thread
    virtual bool suspend();

    // main worker proc
    // MUST be overridden
    virtual int run();

    // sleep in ms
    static void sleep( int ms );
    // wait for event (wait time given in milliseconds or waitForever constant)
    // returns true on success (=>event set)
    static bool wait( Event &ev, int ms = -1 );
};

// get millisecond counter
int getMillisec();
